import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class CosmosCreator extends PApplet {

//// COSMOS CREATOR
//// Created by Virtue One \u00a9 2015
//// Description: Helps user create outer-space backgrounds. 

boolean exitOn;
boolean darkenScreen;
//cursors
PImage cursorImg;
boolean buttonOverColor;
String appTitle;
//controls
boolean colorGrab;
boolean mouseActive;
boolean mouseRightActive;
//boolean ctrlState;
int keyCount; // for controlling unwanted keystrokes
//tools
int toolSize;
int toolSpread;
boolean toolFxPlanetsOn;
boolean toolFxNoiseOn;
boolean toolFxBlurOn;
boolean toolBrushAOn;
boolean toolBrushADarkOn;
boolean toolBrushBOn;
boolean toolBrushCOn;
boolean toolBrushDOn;
boolean flareOn;
//images
PImage imageSparkle;
PImage[] imgFlares;
float[] flareDistances;
float[] flareSizes;
// screen image buffers
PGraphics imageBuffer;
PGraphics imageUndo;
// for gui
boolean showControls;
int buttonSize = 64; 
int buttonSizeSmaller = 40;
//button states
boolean buttonBrushA;
boolean buttonBrushB;
boolean buttonBrushC;
boolean buttonBrushD;
//button images
PImage imageButtonCover;
PImage imageColorBar;
PImage imageButtons;
PImage imageButtonTick;
String buttonDescription;
String buttonTitle;
//general
PFont fontArial;
int colorchoice;
String imageFormat;
int stagew;
int stageh;
int sourcex;
int sourcey;
int c;
int bggradA;
int bggradB;
int bggradC;
int randStemColorFavorRand;

//planets
PImage[] imagePlanets;
int planetNum;

public void setup() {
  //settings
  String lines[] = loadStrings("data/settings.txt");
  String[] thisLineParts; 
  for (int i = 0; i < lines.length; i++) {
    String thisLine = lines[i];
    thisLineParts = split(thisLine, ':');
    if (i==0) stagew = PApplet.parseInt(thisLineParts[1]); 
    if (i==1) stageh = PApplet.parseInt(thisLineParts[1]); 
    if (i==2) imageFormat = thisLineParts[1];
  }
  if (lines.length==0) {
    stagew = 800;
    stageh = 550;
    imageFormat = "png";
  }
  if (stageh<550) stageh = 550;
  if (stagew<360) stagew = 360;
  size(stagew,stageh,P2D);
  ////default vars
  //stagew = 1024;
  //stageh = 600;
  sourcex = stagew/2;
  sourcey = stageh/2;
  appTitle = "Cosmos Creator (v1.0d)";
  frame.setTitle(appTitle);
  
  PGraphics iconPG = createGraphics(64, 64, JAVA2D);
  PImage iconIMG = loadImage("data/icon.png");
  iconPG.beginDraw();
  iconPG.loadPixels();
  iconIMG.loadPixels();
  for (int i=0;i<iconIMG.pixels.length;i++) {
    iconPG.pixels[i] = iconIMG.pixels[i];
  }
  iconPG.updatePixels();
  iconPG.endDraw();
  frame.setIconImage(iconPG.image);
  
  background(0,255);
  strokeWeight(0);
  noStroke();
  ellipseMode(RADIUS); 
  textMode(SHAPE);
  fontArial = createFont("Arial", 20, false);
  textFont(fontArial);
  frameRate(120);  
  buttonTitle = "";
  buttonDescription = "";
  colorchoice = color(50,10,150,255); 
  toolSize = 30;
  toolSpread = 50;
  showControls = false;
  mouseActive = false;
  buttonBrushA = true;
  colorGrab = false;
  buttonOverColor = false;
  planetNum = 0;
  flareOn = false;
  exitOn = false;
  
  setCursor();
  
  // for sparkle brush
  imageSparkle = createImage(100, 100, ARGB);
  imageSparkle = loadImage("data/textures/sparkle.png");
  
  // default buffer/screen
  imageBuffer = createGraphics(stagew,stageh,JAVA2D);
  imageBuffer.beginDraw();
  imageBuffer.strokeWeight(0);
  imageBuffer.noStroke();
  imageBuffer.ellipseMode(RADIUS);
  imageBuffer.background(0,255);  
  imageBuffer.endDraw();
  
  loadGui();
  paintBg(1);
}

public void draw() {
  
  if (showControls) {
    drawGui();    
  } else {
    if (!exitOn && !darkenScreen) {
      int DestX = 0;
      int DestY = 0;
      if (mouseRightActive) {
        DestX = PApplet.parseInt(random(stagew));
        DestY = PApplet.parseInt(random(stageh));
      } else {
        DestX = mouseX;
        DestY = mouseY;
      }
      if (toolBrushAOn) drawClouds(DestX,DestY,random(1,4),PApplet.parseInt(random(toolSize/10+2,toolSize/10+10)),PApplet.parseInt(random(toolSpread*100-500,toolSpread*100+500)),colorchoice,true);
      //if (toolBrushADarkOn) drawClouds(mouseX,mouseY,random(1,4),int(random(toolSize/200+1,toolSize/200+8)),int(random(1000,5000)),colorchoice,false);   
      if (toolBrushADarkOn) drawClouds(mouseX,mouseY,random(1,4),PApplet.parseInt(random(toolSize/5+1,toolSize/5+10)),PApplet.parseInt(random(toolSpread*100-100,toolSpread*100+100)),colorchoice,false);
      if (toolBrushBOn) paintSparkles(DestX,DestY,toolSize,toolSpread);
      if (toolBrushCOn) paintHoles(DestX,DestY,toolSize,toolSpread);
      if (toolBrushDOn) {
        if (mouseRightActive) {
          for(int n=0;n<10;n++) {  
            DestX = PApplet.parseInt(random(stagew));
            DestY = PApplet.parseInt(random(stageh));
            paintStars(DestX,DestY,toolSize,toolSpread);
          }
        } else {
          for(int n=0;n<5;n++) {
            DestX = mouseX;
            DestY = mouseY;
            paintStars(DestX,DestY,toolSize,toolSpread);
          }
        }
      }
    }
    
    drawScreen();

    if (flareOn) {  
      drawFlare();
    }
    if (toolFxPlanetsOn) {
      drawPlanet();
    }
    
    if (!mouseActive) {
      image(imageButtonTick,0,0);
      
      fill(0,100);
      rect(28,4,14,14);
      fill(colorchoice,255);
      rect(30,6,10,10);
      
      if (darkenScreen || exitOn) {
        fill(0,150);
        rect(0,0,stagew,stageh);
        if (exitOn) {
          textSize(18);
          fill(255,100);
          text("LEFT Click to Exit. RIGHT Click to Cancel.",stagew/2-150,stageh/2);
        }
      }
      
      if (buttonOver(1,1,24,24)) {
        showControls=true;
        setCursor();
      }

      textSize(14);
      fill(255,60);
      text("Size: "+toolSize+" Spread: "+toolSpread, 49, 18);
      fill(0,100);
      text("Size: "+toolSize+" Spread: "+toolSpread, 48, 17);
    }
    
  }
  
}

public void setCursor() {
  if (showControls==true) {
    if (buttonOverColor==true) {
      cursorImg = loadImage("data/cursors/cursor_default.png");
      cursor(cursorImg,cursorImg.width/2,cursorImg.height/2);
    } else {
      cursor(ARROW);
    }
  } else {
    if (colorGrab==true || buttonOverColor==true) {
      cursorImg = loadImage("data/cursors/cursor_default.png");
      cursor(cursorImg,cursorImg.width/2,cursorImg.height/2);
    } else {
      if (flareOn || toolFxBlurOn || toolFxNoiseOn) {
        cursor(ARROW);
      } else {
        cursorImg = loadImage("data/cursors/cursor_brush.png");
        cursor(cursorImg,cursorImg.width/2,cursorImg.height/2);
      }
    }
  }     
}

public void loadFlare() {
  File dirFlares = new File(dataPath("/textures/flare"));
  String[] flareFileList = dirFlares.list();
  imgFlares = new PImage[flareFileList.length];
  flareDistances = new float[flareFileList.length];
  flareSizes = new float[flareFileList.length];
  float multiplier = 0;
  for (int i = 0; i < flareFileList.length; i++) {
    imgFlares[i] = createImage(150, 150, ARGB);
    imgFlares[i] = loadImage("data/textures/flare/"+flareFileList[i]);
    multiplier += random(0.15f,1.0f)*i/2;
    flareDistances[i] = multiplier;
    flareSizes[i] = random(2,100);
  }
}

public void drawFlare() {
  blendMode(ADD);
  int size = 0;
  int posX = 0;
  int posY = 0;
  
  //// flares
  for (int i = 0; i < imgFlares.length; i++) {
    int alpha = 0;
    size = PApplet.parseInt(Math.abs((sourcex-mouseX)*flareDistances[i])/50)+PApplet.parseInt(flareSizes[i]);
    size = constrain(size,5,500);
    posX = sourcex + PApplet.parseInt(((sourcex-mouseX))*flareDistances[i]/2)-size/2;
    posY = sourcey + PApplet.parseInt(((sourcey-mouseY))*flareDistances[i]/2)-size/2;
    //// burst
    if (i==0) {
      alpha = Math.abs(sourcex-mouseX)+60;
      size = PApplet.parseInt(Math.abs((sourcex-mouseX)*5));
      size = constrain(size,300,1000);
      posX = mouseX-size/2;
      posY = mouseY-size/2;
    }
    if (i==1) {
      size = PApplet.parseInt((sourcex-mouseX)/2);
      posX = sourcex + PApplet.parseInt((sourcex-mouseX)*-1.68f)-size/2;
      posY = sourcey + PApplet.parseInt((sourcey-mouseY)*-1.68f)-size/2;
    }
    if (i==2) {
      size = PApplet.parseInt((sourcex-mouseX)/4);
      posX = sourcex + PApplet.parseInt((sourcex-mouseX)*-0.8f)-size/2;
      posY = sourcey + PApplet.parseInt((sourcey-mouseY)*-0.8f)-size/2;
    }
    if (i==3) {
      //size = int(Math.abs(((sourcex-mouseX*2)/2))/5);
      size = PApplet.parseInt((sourcex-mouseX)/1.5f);
      posX = sourcex + PApplet.parseInt((sourcex-mouseX)*-1.4f)-size/2;
      posY = sourcey + PApplet.parseInt((sourcey-mouseY)*-1.4f)-size/2;
    }
    if (alpha==0) {
      alpha = Math.abs(sourcex-mouseX);
    }
    alpha = constrain(alpha,50,255);
    tint(255,alpha); // alpha
    image(imgFlares[i],posX,posY,size,size);  //draw flare
  }
  noTint();
  blendMode(BLEND);
}

public void paintBg(int direction) {
  imageBuffer.beginDraw();
  imageBuffer.background(0);
  randStemColorFavorRand = PApplet.parseInt(random(1,4));
  bggradA = PApplet.parseInt(random(20,200));
  bggradB = PApplet.parseInt(random(20,200));
  bggradC = PApplet.parseInt(random(20,200));
  int thisColor = color(0,0,0);
  if (randStemColorFavorRand==1) thisColor = color(random(bggradA-50,bggradA+50),bggradB,bggradC);
  if (randStemColorFavorRand==2) thisColor = color(bggradA,random(bggradB-50,bggradB+50),bggradC);
  if (randStemColorFavorRand==3) thisColor = color(bggradA,bggradB,random(bggradC-50,bggradC+50));
  for (int i = 0; i < stagew; i++) {
    for (int j = 0; j < stageh; j++) {
      if (direction==1) c = color(red(thisColor)-j/5-i/10, green(thisColor)+j/15-i/5, blue(thisColor)+j/5-i/5,255);
      if (direction==2) c = color(bggradA-100+j/5, bggradB-100+j/5, bggradC-100+j/100,255);
      //if (direction==3) c = color(bggradA-20+i/5, bggradB-20+i/5, bggradC+i/50,255);
      if (direction==3) c = color(bggradA-20+i/5, bggradB-20+i/50, bggradC-i/50,255); //updated v1d
      imageBuffer.set(i, j, c);
    }
  }
  colorchoice = thisColor;
  imageBuffer.endDraw();
}

public void loadPlanets() {
  //planets
  File dirPlanets = new File(dataPath("/textures/planets"));
  String[] planetFileList = dirPlanets.list();
  imagePlanets = new PImage[planetFileList.length];
  for (int i = 0; i < planetFileList.length; i++) {
    imagePlanets[i] = createImage(576, 576, ARGB);
    imagePlanets[i] = loadImage("data/textures/planets/"+planetFileList[i]);
  }
}
public void paintPlanet() {
  int planetSize = toolSize*10;
  planetSize = constrain(planetSize,5,1000);
  int planetAlpha = PApplet.parseInt(toolSpread * 2.5f);
  planetAlpha = constrain(planetAlpha,5,255);  
  imageBuffer.beginDraw();
  imageBuffer.tint(255,planetAlpha);
  imageBuffer.image(imagePlanets[planetNum],mouseX-planetSize/2,mouseY-planetSize/2,planetSize,planetSize);
  imageBuffer.tint(255,255);
  imageBuffer.endDraw();
}
public void drawPlanet() {
  int planetSize = toolSize*10;
  planetSize = constrain(planetSize,5,1000);
  //imageBuffer.image(imagePlanets[planetNum],mouseX-planetSize/2,mouseY-planetSize/2,planetSize,planetSize);
  int planetAlpha = PApplet.parseInt(toolSpread * 2.5f);
  planetAlpha = constrain(planetAlpha,5,255);
  tint(255,planetAlpha);
  image(imagePlanets[planetNum],mouseX-planetSize/2,mouseY-planetSize/2,planetSize,planetSize);
  tint(255,255);
}
public void paintNoise() {
  imageBuffer.beginDraw();
  for (int i = 0; i < stagew; i++) {
    for (int j = 0; j < stageh; j++) {
      int baseColor = imageBuffer.get(i, j);
      int noiseColor = color(PApplet.parseInt(random(red(baseColor)-5,red(baseColor)+6)),PApplet.parseInt(random(green(baseColor)-5,green(baseColor+6))),PApplet.parseInt(random(blue(baseColor)-5,blue(baseColor)+6)),255);
      imageBuffer.set(i, j, noiseColor);
    }
  }
  imageBuffer.endDraw();
}

public void paintHoles(int x, int y, int size, int spread) {
  int holeSize = PApplet.parseInt(random(size-10,size+11));
  int randPosX = PApplet.parseInt(random(x-spread,x+spread));
  int randPosY = PApplet.parseInt(random(y-spread,y+spread));
  drawGradient(randPosX,randPosY,holeSize,colorchoice,1.0f,100);
}

public void paintSparkles(int x, int y, int size, int spread) {
  int sparkleSize = PApplet.parseInt(random(size*2-size,size*2+size));
  int randPosX = PApplet.parseInt(random(x-spread,x+spread)-sparkleSize/2);
  int randPosY = PApplet.parseInt(random(y-spread,y+spread)-sparkleSize/2);
  drawGradient(randPosX+sparkleSize/2,randPosY+sparkleSize/2,sparkleSize/4,colorchoice,random(0.4f,0.5f),200);
  imageBuffer.beginDraw();
  imageBuffer.image(imageSparkle,randPosX,randPosY,sparkleSize,sparkleSize);
  imageBuffer.endDraw();
}

public void paintStars(int x, int y, int size, int spread) {
    imageBuffer.beginDraw();
    int starSizeBias = PApplet.parseInt(random(1,4));
    int starAlpha = PApplet.parseInt(random(size,200));
    int randPosX = PApplet.parseInt(random(x-spread,x+spread));
    int randPosY = PApplet.parseInt(random(y-spread,y+spread));
    int randomGrad = PApplet.parseInt(random(1,10));
    if (randomGrad==1) {
      int circSize = PApplet.parseInt(random(1,toolSize/5));
      starAlpha = 200;
      float circDiffuse = random(0.1f,0.5f);
      drawGradient(randPosX,randPosY,circSize+15,colorchoice,circDiffuse,PApplet.parseInt(random(25,100)));
    }
    int randomBias = PApplet.parseInt(random(1,16));
    if (randomBias==1) {
      randPosX = PApplet.parseInt(random(x-spread,x+spread));
      randPosY = PApplet.parseInt(random(y-spread,y+spread));
      drawGradient(randPosX,randPosY,5,color(255,255),10.0f,255);
    }  
    if (starSizeBias==1) {
      starAlpha = starAlpha/2;
      imageBuffer.fill(255,255,255,starAlpha);
      imageBuffer.rect(randPosX,randPosY,1,1);
    } else {
      imageBuffer.fill(255,255,255,starAlpha);
      imageBuffer.ellipse(randPosX,randPosY,1,1);
    }
    imageBuffer.endDraw();
}

public void paintStarCluster(int x, int y, int circSize) {
  imageBuffer.beginDraw();  
  float circDiffuse = random(0.2f,0.5f); 
  int starRadialColor = color(255,255,255);
  int randPosX = PApplet.parseInt(random(x-(PApplet.parseInt(random(5,50))),x+(PApplet.parseInt(random(5,50)))));
  int randPosY = PApplet.parseInt(random(y-(PApplet.parseInt(random(5,50))),y+(PApplet.parseInt(random(5,50))))); 
  // note: 'randStemColorFavorRand' var created in paintBg()
  if (randStemColorFavorRand==1) starRadialColor = color(random(bggradA-50,bggradA+50),bggradB,bggradC);
  if (randStemColorFavorRand==2) starRadialColor = color(bggradA,random(bggradB-50,bggradB+50),bggradC);
  if (randStemColorFavorRand==3) starRadialColor = color(bggradA,bggradB,random(bggradC-50,bggradC+50));
  drawGradient(randPosX,randPosY,circSize+10,starRadialColor,circDiffuse,PApplet.parseInt(random(25,50)));
  imageBuffer.fill(255,255,255,200);
  imageBuffer.ellipse(randPosX,randPosY,1,1);  
  imageBuffer.endDraw();
}

public void paintCluster(int x, int y, int circSize, int clustercolor) {
  imageBuffer.beginDraw();  
  float circDiffuse = random(0.1f,1.0f); 
  drawGradient(x,y,circSize+15,clustercolor,circDiffuse,PApplet.parseInt(random(25,50)));
  imageBuffer.fill(255,255,255,200);
  imageBuffer.ellipse(x,y,1,1);
  for (int n=0;n<PApplet.parseInt(random(1,4));n++) {
    int sizeB = PApplet.parseInt(random(1,2));
    imageBuffer.fill(255,255,255,100);
    imageBuffer.ellipse(PApplet.parseInt(random(x-(PApplet.parseInt(random(5,20))),x+(PApplet.parseInt(random(5,20))))),PApplet.parseInt(random(y-(PApplet.parseInt(random(5,20))),y+(PApplet.parseInt(random(5,20))))),sizeB,sizeB);
  }
  imageBuffer.endDraw();
}

public void paintRadials(int radcolor, int rangeA, int rangeB, int limit) {
  //random circs
  for (int n=0;n<limit;n++) {
    int circSize = PApplet.parseInt(random(rangeA,rangeB));
    float circDiffuse = random(0.1f,1.0f);
    int randPosX = PApplet.parseInt(random(stagew));
    int randPosY = PApplet.parseInt(random(stageh)); 
    drawGradient(randPosX,randPosY,circSize,radcolor,circDiffuse,75);
  }
}

public void drawGradient(float x, float y, int radius, int colorRGB, float diffuseAmnt, float Alpha) {
  imageBuffer.beginDraw();
  // radial gradient
  float h = 0;
  for (int r = radius; r > 0; --r) {
    h = (h + diffuseAmnt) % 100;
    float alpha = h % 100;
    if (alpha>100) alpha = 100;
    if (alpha > Alpha) alpha = Alpha;
    imageBuffer.fill(red(colorRGB), green(colorRGB), blue(colorRGB), alpha);
    imageBuffer.ellipse(x, y, r, r);
  }
  imageBuffer.endDraw();
}


public void drawClouds(int X, int Y, float bitSpeed, int bitsize, int limit, int cloudcolor, boolean clusterbool) {
  imageBuffer.beginDraw(); 
  int x = X;
  int y = Y;
  float rotation = random(-180,180);
  float rotationInRadians;
  int destx;
  int desty;
  int bitsizecount = 0;
  int colorstep = 0;
  int clustercount = 0;
  
  for (int i=0;i<limit;i++) {
    clustercount++;
    bitsizecount++;
    colorstep = colorstep + 1;
    rotation = random(-180,180); // sweet swirly effect.
    rotationInRadians = (rotation * PI) / 180.0f; //rotationInRadians = radians(rotation); // alt method
    destx = PApplet.parseInt((bitSpeed) * cos(rotationInRadians)); // pushes bit in angle direction.
    desty = PApplet.parseInt((bitSpeed) * sin(rotationInRadians));
    x = x+destx;
    y = y+desty;
    int colorchoiceAlpha = colorstep/50;
    if (colorchoiceAlpha>50) colorchoiceAlpha = 50;
    //int colorstepLighten = colorstep*2;
    int colorstepControl = colorstep/75;
    int colorstepControlLess = colorstepControl/5;
    int thiscolor = color(red(cloudcolor),green(cloudcolor),blue(cloudcolor),colorchoiceAlpha);
    // bitsize control
    if (bitsizecount>=100) {
      bitsizecount = 0;
      bitsize = bitsize - 1;
      if (bitsize < 1) bitsize = 1;
    }    
    if (clusterbool==true) {
      if (clustercount>=PApplet.parseInt(random(500,1000))) {
          // this paints star clusters around cloud/debris points
          int gradPosX = PApplet.parseInt(random(x-20,x+20));
          int gradPosY = PApplet.parseInt(random(y-20,y+20));
          paintCluster(gradPosX,gradPosY,bitsize*3,cloudcolor);
          clustercount = 0;
      }
      if (randStemColorFavorRand==1) imageBuffer.fill(red(thiscolor)-colorstepControl,green(thiscolor)-colorstepControlLess,blue(thiscolor)-colorstepControlLess,colorchoiceAlpha);
      if (randStemColorFavorRand==2) imageBuffer.fill(red(thiscolor)-colorstepControlLess,green(thiscolor)-colorstepControl,blue(thiscolor)-colorstepControlLess,colorchoiceAlpha);
      if (randStemColorFavorRand==3) imageBuffer.fill(red(thiscolor)-colorstepControlLess,green(thiscolor)-colorstepControlLess,blue(thiscolor)-colorstepControl,colorchoiceAlpha);
    } else {
      imageBuffer.fill(cloudcolor,colorchoiceAlpha);
    }
    
    imageBuffer.ellipse(x,y,bitsize,bitsize);  //draw bit
    
  } 
  imageBuffer.endDraw();
}



public boolean buttonOver(int x, int y, int w, int h)  {
  if (mouseX >= x && mouseX <= x+w && 
      mouseY >= y && mouseY <= y+h) {
    return true;
  } else {
    return false;
  }
}

public void loadGui() {
  imageButtonCover = createImage(64, 64, ARGB);
  imageButtonCover = loadImage("data/gui/button_cover.png");
  imageButtons = createImage(262, 244, ARGB);
  imageButtons = loadImage("data/gui/buttons.png");
  imageColorBar = createImage(300, 64, ARGB);
  imageColorBar = loadImage("data/gui/colorbar.png");
  imageButtonTick = createImage(16, 16, ARGB);
  imageButtonTick = loadImage("data/gui/button_tick.png");
}

public void drawGui() {

  drawScreen(); // redraw screen, other alpha bleed to solid. 
  
  fill(0,0,0,128);
  rect(0,0,360,550);
  image(imageButtons,32,32);  
  image(imageColorBar,32,459);
  fill(colorchoice);
  rect(319,510,11,11);
 
  //// BUTTON ROLL OVERS...
  //color over
    if (buttonOver(32,459,299,63)) {
      colorchoice = get(mouseX,mouseY);
      buttonOverColor = true;
      setCursor();
    } else {
      buttonOverColor = false;
      setCursor();
    }
  // covers - clouds
    if (buttonOver(32,116,buttonSize,buttonSize)) {
      image(imageButtonCover,32,116);
      buttonTitle = "CLOUDS";
      buttonDescription = "LEFT click to paint. \nRIGHT click to paint without clusters.";
    } else {
      //image(imageButtonBrushA,32,116);
    }  
    if (buttonOver(98,116,buttonSize,buttonSize)) {
      image(imageButtonCover,98,116);
      buttonTitle = "SPARKLES";
      buttonDescription = "LEFT click to paint. \nRIGHT to generate randomly.";
    }
    if (buttonOver(164,116,buttonSize,buttonSize)) {
      image(imageButtonCover,164,116);
      buttonTitle = "HOLES";
      buttonDescription = "LEFT click to paint. \nRIGHT to generate randomly.";
    }
    if (buttonOver(230,116,buttonSize,buttonSize)) {
      image(imageButtonCover,230,116);
      buttonTitle = "STARS";
      buttonDescription = "LEFT click to paint. \nRIGHT to generate randomly.";
    }
    
  // fx
    if (buttonOver(32,200,buttonSize,buttonSize)) {
      image(imageButtonCover,32,200);
      buttonTitle = "BLUR";
      buttonDescription = "LEFT Mouse Button for soft blur. \nRIGHT for strong blur.";
    }  
    if (buttonOver(98,200,buttonSize,buttonSize)) {
      image(imageButtonCover,98,200);
      buttonTitle = "LENS FLARE";
      buttonDescription = "Click LEFT Mouse Button to PLACE. \nRIGHT to set Center Point.";
    }
    if (buttonOver(164,200,buttonSize,buttonSize)) {
      image(imageButtonCover,164,200);
      buttonTitle = "NOISE";
      buttonDescription = "Click LEFT Mouse Button to paint noise.";
    }
    if (buttonOver(230,200,buttonSize,buttonSize)) {
      image(imageButtonCover,230,200);
      buttonTitle = "PLANETS";
      buttonDescription = "LEFT click to paint. \nRIGHT to cycle through images.";
    }
    
  // bg
    if (buttonOver(32,284,buttonSize,buttonSize)) {
      image(imageButtonCover,32,284);
      buttonTitle = "BACKGROUND";
      buttonDescription = "Generates random color gradient for background \n(split).";
    }  
    if (buttonOver(98,284,buttonSize,buttonSize)) {
      image(imageButtonCover,98,284);
      buttonTitle = "BACKGROUND";
      buttonDescription = "Generates random color gradient for background \n(horizontal).";
    }
    if (buttonOver(164,284,buttonSize,buttonSize)) {
      image(imageButtonCover,164,284);
      buttonTitle = "BACKGROUND";
      buttonDescription = "Generates random color gradient for background \n(vertical).";
    }
    if (buttonOver(230,284,buttonSize,buttonSize)) {
      image(imageButtonCover,230,284);
      buttonTitle = "SCENE";
      buttonDescription = "Generates fully random scene - clouds, stars, etc. \n(may take several moments to process)";
    }
    
    // app controls
    if (buttonOver(32,32,buttonSizeSmaller,buttonSizeSmaller)) {
      image(imageButtonCover,32,32,buttonSizeSmaller,buttonSizeSmaller);
      buttonTitle = "EXIT";
      buttonDescription = "";
    }
    if (buttonOver(80,32,buttonSizeSmaller,buttonSizeSmaller)) {
      image(imageButtonCover,80,32,buttonSizeSmaller,buttonSizeSmaller);
      buttonTitle = "OPEN";
      buttonDescription = "";
    }
    if (buttonOver(128,32,buttonSizeSmaller,buttonSizeSmaller)) {
      image(imageButtonCover,128,32,buttonSizeSmaller,buttonSizeSmaller);
      buttonTitle = "SAVE";
      buttonDescription = "";
    }
    if (buttonOver(176,32,buttonSizeSmaller,buttonSizeSmaller)) {
      image(imageButtonCover,176,32,buttonSizeSmaller,buttonSizeSmaller);
      buttonTitle = "UNDO/REDO";
      buttonDescription = "";
    }    
    
    
    /// gui text
    fill(255,150);
    //if (buttonBrushA || buttonBrushB || buttonBrushC || buttonBrushD) {
      //textSize(16);
      //text("Size: "+toolSize+" Spread: "+toolSpread, 32, 442);
    //}
    textSize(12);
    text(buttonDescription, 32, 400);
    textSize(18);
    text(buttonTitle, 32, 380);
    
}

public void drawScreen() {
  //background(0,255);
  image(imageBuffer,0,0); 
}







/// controls

public void mousePressed() {
  mouseActive = true;
  if (mouseButton == LEFT) {
    if (colorGrab==true) {
      colorchoice = get(mouseX,mouseY);
      colorGrab = false;
    } else {
      if (showControls==false) {
        if (flareOn) {
          imageUndo = saveBufferPG(); // save undo
          drawScreen();
          drawFlare();
          imageBuffer.beginDraw();
          imageBuffer.image(saveImg(),0,0);
          imageBuffer.endDraw();
        } else {
          imageUndo = saveBufferPG(); // save undo
          if (buttonBrushA) toolBrushAOn = true;
          if (buttonBrushB) toolBrushBOn = true;
          if (buttonBrushC) toolBrushCOn = true;
          if (buttonBrushD) toolBrushDOn = true;
          if (toolFxBlurOn) { 
            imageBuffer.beginDraw();
            imageBuffer.filter(BLUR,2);
            imageBuffer.endDraw();
          }
          if (toolFxNoiseOn) {
            paintNoise();
          }
          if (toolFxPlanetsOn) {
            paintPlanet();
          }
        }
      }
    }
  }
  if (mouseButton == RIGHT) {
    mouseRightActive = true;
    if (flareOn) {
      loadFlare();
      sourcex = mouseX;
      sourcey = mouseY;
    } else {
      imageUndo = saveBufferPG();  // save undo
      if (buttonBrushA) toolBrushADarkOn = true;
      if (buttonBrushB) toolBrushBOn = true;
      if (buttonBrushC) toolBrushCOn = true;
      if (buttonBrushD) toolBrushDOn = true;
      if (toolFxBlurOn) {
        imageBuffer.beginDraw();
        imageBuffer.filter(BLUR,10);
        imageBuffer.endDraw();
      }
      if (toolFxPlanetsOn) {
        planetNum++;
        if (planetNum>imagePlanets.length-1) planetNum = 0;
      }
    }
  }
  setCursor();
}
public void mouseReleased() {
  mouseActive = false;
  if (mouseButton == LEFT) {
    if (exitOn) exit();
    if (buttonBrushA) toolBrushAOn = false;
    if (buttonBrushB) toolBrushBOn = false;
    if (buttonBrushC) toolBrushCOn = false;
    if (buttonBrushD) toolBrushDOn = false;
    if (showControls==true) {
      guiActions();
    }
  }
  if (mouseButton == RIGHT) {
    mouseRightActive = false;
    if (exitOn) exitOn = false;
    if (buttonBrushA) toolBrushADarkOn = false;
    if (buttonBrushB) toolBrushBOn = false;
    if (buttonBrushC) toolBrushCOn = false;
    if (buttonBrushD) toolBrushDOn = false;
  }
  setCursor();
}
    
public void keyPressed() {
  if (key == ESC) key=0; // trap ESC so it doesn't quit
  //// SPREAD and SIZE key input
  if (key=='w') {
    if (toolSize<30) {
      if (toolSize<10) {
        toolSize = toolSize + 1;
      } else {
        toolSize = toolSize + 2;
      }
    } else {
      toolSize = toolSize + 10;
    }
    toolSize = constrain(toolSize,1,100);
  }
  if (key=='s') {
    if (toolSize>1) {
      if (toolSize<40) {
        if (toolSize<10) {
          toolSize = toolSize - 1;
        } else {
          toolSize = toolSize - 2;
        }
      } else {
        toolSize = toolSize - 10;
      }
    }
    toolSize = constrain(toolSize,1,100);
  }
  if (key=='a') {
    if (toolSpread>1) {
      if (toolSpread<30) {
        if (toolSpread<10) {
          toolSpread = toolSpread - 1;
        } else {
          toolSpread = toolSpread - 2;
        }
      } else {
        toolSpread = toolSpread - 10;
      }
    }
    toolSpread = constrain(toolSpread,1,100);
  }
  if (key=='d') {
    if (toolSpread<40) {
      if (toolSpread<10) {
        toolSpread = toolSpread + 1;
      } else {
        toolSpread = toolSpread + 2;
      }
    } else {
      toolSpread = toolSpread + 10;
    }
    toolSpread = constrain(toolSpread,1,100);
  }
  
  if (key=='x') {
    keyCount++; // control var
    if(keyCount==1) {
      showControls=!showControls;
      setCursor();
    }
  }

  if (showControls==false) {
    if (key=='c') {
      if (showControls==false) {
        colorGrab = true;
        setCursor();
      }
    }  
  }
  
  //// method for Ctrl key use...
  //if (key == CODED) {
    //if (keyCode == CONTROL) {
      //ctrlState = true;
    //}
  //} else {
    //if (ctrlState && keyCode == 81) { // 'q' key
      //exit();
    //}
  //}
  
}

public void keyReleased() {
  if (key=='q') {
    if (exitOn) exit();
    if (!exitOn) exitOn = true;
  }
  if (key=='c') {
    colorGrab = false;
    setCursor();
  }
  if (key=='z') {
    loadUndo();
  }  
  if (key=='x') {
    keyCount = 0;
    setCursor();
    //guiActions();
  }
  if (key=='e') {
    saveBufferPG().save("exports/"+frameCount+"."+imageFormat);
  }
  if (showControls==false) {
    if (key=='f') {
      darkenScreen = true;
      drawScreen();
      selectInput("Select a file:", "fileSelected");
    }
  }
}



public void guiActions() {
    showControls=false;
    keyCount = 0;
    buttonOverColor = false;
    
    //// BUTTON ACTIONS (in order)
    
    if (buttonOver(32,116,buttonSize,buttonSize)) {
      buttonBrushA = true;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }
    if (buttonOver(98,116,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = true;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }
    if (buttonOver(164,116,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = true;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }
    if (buttonOver(230,116,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = true;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }
    
    // FX
    if (buttonOver(32,200,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = true;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }  
    if (buttonOver(98,200,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = true;
      loadFlare();
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = false;
    }
    if (buttonOver(164,200,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = true;
      toolFxPlanetsOn = false;
    }
    if (buttonOver(230,200,buttonSize,buttonSize)) {
      buttonBrushA = false;
      buttonBrushB = false;
      buttonBrushC = false;
      buttonBrushD = false;
      //fx
      flareOn = false;
      toolFxBlurOn = false;
      toolFxNoiseOn = false;
      toolFxPlanetsOn = true;
      loadPlanets();
    }
    
    // BG
    if (buttonOver(32,284,buttonSize,buttonSize)) {
      imageUndo = saveBufferPG();
      paintBg(1);
    }
    if (buttonOver(98,284,buttonSize,buttonSize)) {
      imageUndo = saveBufferPG();
      paintBg(2);
    }
    if (buttonOver(164,284,buttonSize,buttonSize)) {
      imageUndo = saveBufferPG();
      paintBg(3);
    }
    if (buttonOver(230,284,buttonSize,buttonSize)) {
      imageUndo = saveBufferPG();
      randomScene();
    }
    
    // app controls
    if (buttonOver(32,32,buttonSizeSmaller,buttonSizeSmaller)) {
      //quit
      exitOn = true;
    }
    if (buttonOver(80,32,buttonSizeSmaller,buttonSizeSmaller)) {
      //open
      darkenScreen = true;
      drawScreen();
      selectInput("Select a file:", "fileSelected");
    }
    if (buttonOver(128,32,buttonSizeSmaller,buttonSizeSmaller)) {
      //save
      saveBufferPG().save("exports/"+frameCount+"."+imageFormat);
    }
    if (buttonOver(176,32,buttonSizeSmaller,buttonSizeSmaller)) {
      //undo
      loadUndo();
    }    
    drawScreen(); //refresh
}



//// OPEN FILE AND SET SCREEN
public void fileSelected(File selection) {
  darkenScreen = false;
  if (selection == null) {
    //println("Window was closed or the user hit cancel.");
  } else {
    imageUndo = saveBufferPG();
    PImage imageLoaded = createImage(stagew, stageh, ARGB);
    imageLoaded = loadImage(selection.getAbsolutePath());
    imageBuffer.beginDraw();
    imageBuffer.image(imageLoaded,0,0,stagew,stageh);
    imageBuffer.endDraw();
    frame.setTitle(appTitle+" ["+selection.getName()+"]");
    //resizeCanvas();  // for future...
  }
}

//// screen save
public PImage saveImg() {
  PImage img = createImage(stagew, stageh, RGB);
  loadPixels();
  img.loadPixels();
  for (int x=0;x<stagew;x++) {
    for (int y=0;y<stageh;y++) {
      if (img.width==stagew) {
        img.pixels[y*img.width+x] = pixels[y*stagew+x];
      }
    }
  }
  img.updatePixels();
  return img;
}
//// buffer save
public PGraphics saveBufferPG() {
  PGraphics img = createGraphics(stagew,stageh,JAVA2D);
  img.beginDraw();
  img.strokeWeight(0);
  img.noStroke();
  img.ellipseMode(RADIUS);
  img.background(0,255);  
  img.endDraw();
  imageBuffer.loadPixels();
  img.loadPixels();
  for (int x=0;x<stagew;x++) {
    for (int y=0;y<stageh;y++) {
      if (img.width==stagew) {
        img.pixels[y*img.width+x] = imageBuffer.pixels[y*stagew+x];
      }
    }
  }
  img.updatePixels();
  return img;
}

public void loadUndo() {
  if (imageUndo!=null) {
    PGraphics img = createGraphics(stagew,stageh,JAVA2D);
    img.beginDraw();
    img.strokeWeight(0);
    img.noStroke();
    img.ellipseMode(RADIUS);
    img.background(0,255);  
    img.endDraw();
    img = imageBuffer;
    imageBuffer = imageUndo;
    imageUndo = img; 
  }
}

//// generates random scene with fx...
public void randomScene() {
  int DestX = 0;
  int DestY = 0;
  int thisToolSize = 10;
  int thisToolSpread = 10;
  paintBg(PApplet.parseInt(random(1,4))); //random bg
  int prevColorchoice = colorchoice;
  //clouds, processed with blur and noise.
  for(int m=0;m<3;m++) {
    if (m!=0) imageBuffer.filter(BLUR,10); // blur fx
    for(int n=0;n<3;n++) {
      thisToolSize = PApplet.parseInt(random(10,100));
      thisToolSpread = PApplet.parseInt(random(10,100));
      DestX = PApplet.parseInt(random(stagew));
      DestY = PApplet.parseInt(random(stageh));
      drawClouds(DestX,DestY,random(1,4),PApplet.parseInt(random(thisToolSize/10+2,thisToolSize/10+10)),PApplet.parseInt(random(thisToolSpread*100-500,thisToolSpread*100+500)),colorchoice,true);
    }
    for(int n=0;n<5;n++) {
      DestX = PApplet.parseInt(random(stagew));
      DestY = PApplet.parseInt(random(stageh));
      thisToolSize = PApplet.parseInt(random(10,100));
      thisToolSpread = PApplet.parseInt(random(10,100));
      //black clouds
      drawClouds(DestX,DestY,random(1,4),PApplet.parseInt(random(thisToolSize/5+1,thisToolSize/5+10)),PApplet.parseInt(random(thisToolSpread*100-100,thisToolSpread*100+100)),color(0,0,0,255),false);
    }
    paintNoise(); 
  }
  // holes
  colorchoice = color(0,0,0,255); //force black color
  for(int n=0;n<10;n++) {
    DestX = PApplet.parseInt(random(stagew));
    DestY = PApplet.parseInt(random(stageh));
    thisToolSize = PApplet.parseInt(random(5,PApplet.parseInt(random(6,60))));
    thisToolSpread = PApplet.parseInt(random(10,100));
    paintHoles(DestX,DestY,thisToolSize,thisToolSpread);
  }   
  colorchoice = prevColorchoice; // revert color choice
  //stars
  for(int n=0;n<500;n++) {  
    DestX = PApplet.parseInt(random(stagew));
    DestY = PApplet.parseInt(random(stageh));
    thisToolSize = PApplet.parseInt(random(5,50));
    thisToolSpread = PApplet.parseInt(random(10,100));
    paintStars(DestX,DestY,thisToolSize,thisToolSpread);
  }
  //sparkles
  for(int n=0;n<15;n++) {
    DestX = PApplet.parseInt(random(stagew));
    DestY = PApplet.parseInt(random(stageh));
    thisToolSize = PApplet.parseInt(random(5,50));
    thisToolSpread = PApplet.parseInt(random(10,100));
    paintSparkles(DestX,DestY,thisToolSize,thisToolSpread);
  }  
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "CosmosCreator" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
